@@ -23,19 +23,45 @@
 #define BABL_LEGAL_ERROR           0.000001
 #define BABL_MAX_COST_VALUE        2000000
 #define BABL_HARD_MAX_PATH_LENGTH  8
+#define BABL_MAX_NAME_LEN          1024
+
+
+typedef struct _FishPathInstrumentation
+{
+  Babl   *fmt_rgba_double;
+  double *test;
+  void   *source;
+  void   *destination;
+  void   *ref_destination;
+  double *destination_rgba_double;
+  double *ref_destination_rgba_double;
+  Babl   *fish_rgba_to_source;
+  Babl   *fish_reference;
+  Babl   *fish_destination_to_rgba;
+  double  reference_cost;
+  int     init_instrumentation_done;
+} FishPathInstrumentation;
+
+typedef struct PathContext {
+  Babl     *fish_path;
+  Babl     *to_format;
+  BablList *current_path;
+} PathContext;
 
 static void
-init_path_instrumentation (Babl *fmt_source,
-                           Babl *fmt_destination);
+init_path_instrumentation (FishPathInstrumentation *fpi,
+                           Babl                    *fmt_source,
+                           Babl                    *fmt_destination);
 
 static void
-destroy_path_instrumentation (void);
+destroy_path_instrumentation (FishPathInstrumentation *fpi);
 
 static void
-get_path_instrumentation (BablList *path,
-                          double   *path_cost,
-                          double   *ref_cost,
-                          double   *path_error);
+get_path_instrumentation (FishPathInstrumentation *fpi,
+                          BablList                *path,
+                          double                  *path_cost,
+                          double                  *ref_cost,
+                          double                  *path_error);
 
 static long
 process_conversion_path (BablList *path,
@@ -44,7 +70,8 @@ process_conversion_path (BablList *path,
                          long     n);
 
 static void
-get_conversion_path (Babl *current_format,
+get_conversion_path (PathContext *pc,
+                     Babl *current_format,
                      int current_length,
                      int max_length);
 
@@ -52,7 +79,8 @@ static double *
 test_create (void);
 
 static char *
-create_name (const Babl *source,
+create_name (char       *buf,
+             const Babl *source,
              const Babl *destination,
              int         is_reference);
 
@@ -111,14 +139,12 @@ static int max_path_length (void)
  * implemented by recursive function get_conversion_path ().
  */
 
-static Babl *fish_path;
-static Babl *to_format;
-static BablList *current_path;
 
 static void
-get_conversion_path (Babl *current_format,
-                     int current_length,
-                     int max_length)
+get_conversion_path (PathContext *pc,
+                     Babl        *current_format,
+                     int          current_length,
+                     int          max_length)
 {
   if (current_length > max_length)
     {
@@ -126,7 +152,7 @@ get_conversion_path (Babl *current_format,
        * depth, let's bail out */
       return;
     }
-  else if ((current_length > 0) && (current_format == to_format))
+  else if ((current_length > 0) && (current_format == pc->to_format))
     {
        /* We have found a candidate path, let's
         * see about it's properties */
@@ -135,37 +161,43 @@ get_conversion_path (Babl *current_format,
       double path_error = 1.0;
       int    i;
 
-      for (i = 0; i < babl_list_size (current_path); i++)
+      for (i = 0; i < babl_list_size (pc->current_path); i++)
         {
-          path_error *= (1.0 + babl_conversion_error ((BablConversion *) current_path->items[i]));
+          path_error *= (1.0 + babl_conversion_error ((BablConversion *) pc->current_path->items[i]));
         }
 
       if (path_error - 1.0 <= legal_error ()) /* check this before the next;
                                                  which does a more accurate
                                                  measurement of the error */
         {
-          get_path_instrumentation (current_path, &path_cost, &ref_cost, &path_error);
+          FishPathInstrumentation fpi;
+          memset (&fpi, 0, sizeof (fpi));
+          get_path_instrumentation (&fpi, pc->current_path, &path_cost, &ref_cost, &path_error);
 
           if ((path_cost < ref_cost) && /* do not use paths that took longer to compute than reference */
-              (path_cost < fish_path->fish_path.cost) &&
+              (path_cost < pc->fish_path->fish_path.cost) &&
               (path_error <= legal_error ()))
             {
               /* We have found the best path so far,
                * let's copy it into our new fish */
-              fish_path->fish_path.cost = path_cost;
-              fish_path->fish.error  = path_error;
-              babl_list_copy (current_path,
-                              fish_path->fish_path.conversion_list);
+              pc->fish_path->fish_path.cost = path_cost;
+              pc->fish_path->fish.error  = path_error;
+              babl_list_copy (pc->current_path,
+                              pc->fish_path->fish_path.conversion_list);
             }
+
+          destroy_path_instrumentation (&fpi);
         }
     }
   else
     {
       /*
-       * Bummer, we have to search deeper... */
+       * Bummer, we have to search deeper... 
+       */
       BablList *list;
       int i;
 
+
       list = current_format->format.from_list;
       if (list)
         {
@@ -180,27 +212,26 @@ get_conversion_path (Babl *current_format,
               if (!next_format->format.visited)
                 {
                   /* next_format is not in the current path, we can pay a visit */
-                  babl_list_insert_last (current_path, next_conversion);
-                  get_conversion_path (next_format, current_length + 1, max_length);
-                  babl_list_remove_last (current_path);
+                  babl_list_insert_last (pc->current_path, next_conversion);
+                  get_conversion_path (pc, next_format, current_length + 1, max_length);
+                  babl_list_remove_last (pc->current_path);
                 }
             }
 
           /* Remove the current format from current path */
           current_format->format.visited = 0;
         }
-      }
+   }
 }
 
-static char buf[1024];
-
 static char *
-create_name (const Babl *source,
+create_name (char       *buf,
+             const Babl *source,
              const Babl *destination,
              int         is_reference)
 {
   /* fish names are intentionally kept short */
-  snprintf (buf, 1024, "%s %p %p", "",
+  snprintf (buf, BABL_MAX_NAME_LEN, "%s %p %p", "",
             source, destination);
   return buf;
 }
@@ -210,9 +241,9 @@ babl_fish_path (const Babl *source,
                 const Babl *destination)
 {
   Babl *babl = NULL;
-  char *name;
+  char name[BABL_MAX_NAME_LEN];
 
-  name = create_name (source, destination, 1);
+  create_name (name, source, destination, 1);
   babl = babl_db_exist_by_name (babl_fish_db (), name);
   if (babl)
     {
@@ -233,20 +264,28 @@ babl_fish_path (const Babl *source,
   babl->fish.destination          = destination;
   babl->fish.processings          = 0;
   babl->fish.pixels               = 0;
-  babl->fish.usecs                = 0;
   babl->fish.error                = BABL_MAX_COST_VALUE;
   babl->fish_path.cost            = BABL_MAX_COST_VALUE;
   babl->fish_path.loss            = BABL_MAX_COST_VALUE;
   babl->fish_path.conversion_list = babl_list_init_with_size (BABL_HARD_MAX_PATH_LENGTH);
 
-  current_path = babl_list_init_with_size (BABL_HARD_MAX_PATH_LENGTH);
-  fish_path = babl;
-  to_format = (Babl *) destination;
+  {
+    PathContext pc;
+    pc.current_path = babl_list_init_with_size (BABL_HARD_MAX_PATH_LENGTH);
+    pc.fish_path = babl;
+    pc.to_format = (Babl *) destination;
 
-  get_conversion_path ((Babl *) source, 0, max_path_length ());
-  destroy_path_instrumentation ();
+    babl_mutex_lock (babl_format_mutex);
+    /* we hold a global lock whilerunning get_conversion_path since
+     * it depends on keeping the various format.visited members in
+     * a consistent state, this code path is not performance critical
+     * since created fishes are cached.
+     */
 
-  babl_list_destroy (current_path);
+    get_conversion_path (&pc, (Babl *) source, 0, max_path_length ());
+    babl_mutex_unlock (babl_format_mutex);
+    babl_list_destroy (pc.current_path);
+  }
 
   if (babl_list_size (babl->fish_path.conversion_list) == 0)
     {
@@ -262,25 +301,99 @@ babl_fish_path (const Babl *source,
   return babl;
 }
 
-long
+static long
 babl_fish_path_process (Babl *babl,
                         void *source,
                         void *destination,
                         long  n)
 {
-  long ret;
-
-  babl_assert (source);
-  babl_assert (destination);
-
-  ret = process_conversion_path (babl->fish_path.conversion_list,
+  return process_conversion_path (babl->fish_path.conversion_list,
                                  source,
                                  destination,
                                  n);
 
+}
+
+
+static long
+babl_fish_process (Babl *babl,
+                   void *source,
+                   void *destination,
+                   long  n)
+{
+  long ret = 0;
+
+  switch (babl->class_type)
+    {
+      case BABL_FISH_REFERENCE:
+        if (babl->fish.source == babl->fish.destination)
+          { /* XXX: we're assuming linear buffers */
+            memcpy (destination, source, n * babl->fish.source->format.bytes_per_pixel);
+            ret = n;
+          }
+        else
+          {
+            ret = babl_fish_reference_process (babl, source, destination, n);
+          }
+        break;
+
+      case BABL_FISH_SIMPLE:
+        if (BABL (babl->fish_simple.conversion)->class_type == BABL_CONVERSION_LINEAR)
+          {
+            ret = babl_conversion_process (BABL (babl->fish_simple.conversion),
+                                           source, destination, n);
+          }
+        else
+          {
+            babl_fatal ("Cannot use a simple fish to process without a linear conversion");
+          }
+        break;
+
+      case BABL_FISH_PATH:
+        ret = babl_fish_path_process (babl, source, destination, n);
+        break;
+
+      default:
+        babl_log ("NYI");
+        ret = -1;
+        break;
+    }
+
   return ret;
 }
 
+long
+babl_process (Babl *babl,
+              void *source,
+              void *destination,
+              long  n)
+{
+  babl_assert (babl);
+  babl_assert (source);
+  babl_assert (destination);
+  babl_assert (BABL_IS_BABL (babl));
+  if (n == 0)
+    return 0;
+  babl_assert (n > 0);
+
+  /* first check if it is a fish since that is out fast path */
+  if (babl->class_type >= BABL_FISH &&
+      babl->class_type <= BABL_FISH_PATH)
+    {
+      babl->fish.processings++;
+      return babl->fish.pixels +=
+             babl_fish_process (babl, source, destination, n);
+    }
+
+  /* matches all conversion classes */
+  if (babl->class_type >= BABL_CONVERSION &&
+      babl->class_type <= BABL_CONVERSION_PLANAR)
+    return babl_conversion_process (babl, source, destination, n);
+
+  babl_fatal ("eek");
+  return -1;
+}
+
 static long
 process_conversion_path (BablList *path,
                          void     *source_buffer,
@@ -352,6 +465,10 @@ test_create (void)
 {
   static double test[sizeof (double) * NUM_TEST_PIXELS * 4];
   int           i, j;
+  static int    done = 0;
+
+  if (done)
+    return test;
 
   /* There is no need to generate the test
    * more times ... */
@@ -370,34 +487,21 @@ test_create (void)
   for (j = 0; j < 16 * 4; i++, j++)
     test [i] = 1.0 + (double) random () / RAND_MAX;
 
+  done = 1;
   return test;
 }
 
-/* FishPath instrumentation */
-
-static Babl   *fmt_rgba_double = NULL;
-static double *test = NULL;
-static void   *source;
-static void   *destination;
-static void   *ref_destination;
-static double *destination_rgba_double;
-static double *ref_destination_rgba_double;
-static Babl   *fish_rgba_to_source;
-static Babl   *fish_reference;
-static Babl   *fish_destination_to_rgba;
-static double  reference_cost;
-static int     init_instrumentation_done = 0;
-
 static void
-init_path_instrumentation (Babl *fmt_source,
-                           Babl *fmt_destination)
+init_path_instrumentation (FishPathInstrumentation *fpi,
+                           Babl                    *fmt_source,
+                           Babl                    *fmt_destination)
 {
   long   ticks_start = 0;
   long   ticks_end   = 0;
 
-  if (!fmt_rgba_double)
+  if (!fpi->fmt_rgba_double)
     {
-      fmt_rgba_double = babl_format_new (
+      fpi->fmt_rgba_double = babl_format_new (
         babl_model ("RGBA"),
         babl_type ("double"),
         babl_component ("R"),
@@ -407,102 +511,105 @@ init_path_instrumentation (Babl *fmt_source,
         NULL);
     }
 
-  if (!test)
-    test = test_create ();
+  if (!fpi->test)
+    fpi->test = test_create (); // <- test_create utiliza var static dentro de la función
 
-  fish_rgba_to_source      = babl_fish_reference (fmt_rgba_double,
+  fpi->fish_rgba_to_source      = babl_fish_reference (fpi->fmt_rgba_double,
                                                   fmt_source);
-  fish_reference           = babl_fish_reference (fmt_source,
+  fpi->fish_reference           = babl_fish_reference (fmt_source,
                                                   fmt_destination);
-  fish_destination_to_rgba = babl_fish_reference (fmt_destination,
-                                                  fmt_rgba_double);
+  fpi->fish_destination_to_rgba = babl_fish_reference (fmt_destination,
+                                                  fpi->fmt_rgba_double);
 
-  source                      = babl_calloc (NUM_TEST_PIXELS,
+  fpi->source                      = babl_calloc (NUM_TEST_PIXELS,
                                              fmt_source->format.bytes_per_pixel);
-  destination                 = babl_calloc (NUM_TEST_PIXELS,
+  fpi->destination                 = babl_calloc (NUM_TEST_PIXELS,
                                              fmt_destination->format.bytes_per_pixel);
-  ref_destination             = babl_calloc (NUM_TEST_PIXELS,
+  fpi->ref_destination             = babl_calloc (NUM_TEST_PIXELS,
                                              fmt_destination->format.bytes_per_pixel);
-  destination_rgba_double     = babl_calloc (NUM_TEST_PIXELS,
-                                             fmt_rgba_double->format.bytes_per_pixel);
-  ref_destination_rgba_double = babl_calloc (NUM_TEST_PIXELS,
-                                             fmt_rgba_double->format.bytes_per_pixel);
+  fpi->destination_rgba_double     = babl_calloc (NUM_TEST_PIXELS,
+                                             fpi->fmt_rgba_double->format.bytes_per_pixel);
+  fpi->ref_destination_rgba_double = babl_calloc (NUM_TEST_PIXELS,
+                                             fpi->fmt_rgba_double->format.bytes_per_pixel);
 
   /* create sourcebuffer from testbuffer in the correct format */
-  babl_process (fish_rgba_to_source,
-                test, source, NUM_TEST_PIXELS);
+  babl_process (fpi->fish_rgba_to_source,
+                fpi->test, fpi->source, NUM_TEST_PIXELS);
 
   /* calculate the reference buffer of how it should be */
   ticks_start = babl_ticks ();
-  babl_process (fish_reference,
-                source, ref_destination, NUM_TEST_PIXELS);
+  babl_process (fpi->fish_reference,
+                fpi->source, fpi->ref_destination, NUM_TEST_PIXELS);
   ticks_end = babl_ticks ();
-  reference_cost = babl_process_cost (ticks_start, ticks_end);
+  fpi->reference_cost = babl_process_cost (ticks_start, ticks_end);
 
   /* transform the reference destination buffer to RGBA */
-  babl_process (fish_destination_to_rgba,
-                ref_destination, ref_destination_rgba_double, NUM_TEST_PIXELS);
+  babl_process (fpi->fish_destination_to_rgba,
+                fpi->ref_destination, fpi->ref_destination_rgba_double, NUM_TEST_PIXELS);
 }
 
 static void
-destroy_path_instrumentation (void)
+destroy_path_instrumentation (FishPathInstrumentation *fpi)
 {
-  if (init_instrumentation_done)
+  if (fpi->init_instrumentation_done)
     {
-      babl_free (source);
-      babl_free (destination);
-      babl_free (destination_rgba_double);
-      babl_free (ref_destination);
-      babl_free (ref_destination_rgba_double);
+      babl_free (fpi->source);
+      babl_free (fpi->destination);
+      babl_free (fpi->destination_rgba_double);
+      babl_free (fpi->ref_destination);
+      babl_free (fpi->ref_destination_rgba_double);
 
       /* nulify the flag for potential new search */
-      init_instrumentation_done = 0;
+      fpi->init_instrumentation_done = 0;
   }
 }
 
 static void
-get_path_instrumentation (BablList *path,
-                          double   *path_cost,
-                          double   *ref_cost,
-                          double   *path_error)
+get_path_instrumentation (FishPathInstrumentation *fpi,
+                          BablList                *path,
+                          double                  *path_cost,
+                          double                  *ref_cost,
+                          double                  *path_error)
 {
   long   ticks_start = 0;
   long   ticks_end   = 0;
 
-  if (!init_instrumentation_done)
+  if (!fpi->init_instrumentation_done)
     {
       /* this initialization can be done only once since the
        * source and destination formats do not change during
        * the search */
       Babl *fmt_source = (Babl *) BABL (babl_list_get_first (path))->conversion.source;
       Babl *fmt_destination = (Babl *) BABL (babl_list_get_last (path))->conversion.destination;
-      init_path_instrumentation (fmt_source, fmt_destination);
-      init_instrumentation_done = 1;
+      init_path_instrumentation (fpi, fmt_source, fmt_destination);
+      fpi->init_instrumentation_done = 1;
     }
 
   /* calculate this path's view of what the result should be */
   ticks_start = babl_ticks ();
-  process_conversion_path (path, source, destination, NUM_TEST_PIXELS);
+  process_conversion_path (path, fpi->source, fpi->destination, NUM_TEST_PIXELS);
   ticks_end = babl_ticks ();
   *path_cost = babl_process_cost (ticks_start, ticks_end);
 
   /* transform the reference and the actual destination buffers to RGBA
    * for comparison with each other
    */
-  babl_process (fish_destination_to_rgba,
-                destination, destination_rgba_double, NUM_TEST_PIXELS);
+  babl_process (fpi->fish_destination_to_rgba,
+                fpi->destination, fpi->destination_rgba_double, NUM_TEST_PIXELS);
 
-  *path_error = babl_rel_avg_error (destination_rgba_double,
-                                    ref_destination_rgba_double,
+  *path_error = babl_rel_avg_error (fpi->destination_rgba_double,
+                                    fpi->ref_destination_rgba_double,
                                     NUM_TEST_PIXELS * 4);
 
-  fish_rgba_to_source->fish.processings--;
-  fish_reference->fish.processings--;
-  fish_destination_to_rgba->fish.processings -= 2;
+#if 0
+  fpi->fish_rgba_to_source->fish.processings--;
+  fpi->fish_reference->fish.processings--;
+  fpi->fish_destination_to_rgba->fish.processings -= 2;
 
-  fish_rgba_to_source->fish.pixels      -= NUM_TEST_PIXELS;
-  fish_reference->fish.pixels           -= NUM_TEST_PIXELS;
-  fish_destination_to_rgba->fish.pixels -= 2 * NUM_TEST_PIXELS;
+  fpi->fish_rgba_to_source->fish.pixels      -= NUM_TEST_PIXELS;
+  fpi->fish_reference->fish.pixels           -= NUM_TEST_PIXELS;
+  fpi->fish_destination_to_rgba->fish.pixels -= 2 * NUM_TEST_PIXELS;
+#endif
 
-  *ref_cost = reference_cost;
+  *ref_cost = fpi->reference_cost;
 }
