@@ -18,6 +18,7 @@
 
 #include "config.h"
 #include "babl-internal.h"
+#include <stddef.h>
 #include <string.h>
 #include <stdarg.h>
 
@@ -137,7 +138,7 @@ babl_fish_get_id (const Babl *source,
   /* value of 'id' will be used as argument for hash function,
    * substraction serves as simple combination of
    * source/destination values. */
-  int id = (int) source - (int) destination;
+  ptrdiff_t id = source - destination;
   /* instances with id 0 won't be inserted into database */
   if (id == 0)
     id = 1;
@@ -210,8 +211,7 @@ babl_fish (const void *source,
         /* In the case of equal source and destination formats
          * we will search through the fish database for reference fish
          * to handle the memcpy */
-        id_htable->find_func = find_memcpy_fish;
-        babl_hash_table_find (id_htable, hashval, (void *) &ffish);
+        babl_hash_table_find (id_htable, hashval, find_memcpy_fish, (void *) &ffish);
       }
     else
       {
@@ -228,8 +228,7 @@ babl_fish (const void *source,
          * insert it into the fish database to indicate non-existent fish
          * path.
          */
-        id_htable->find_func = find_fish_path;
-        babl_hash_table_find (id_htable, hashval, (void *) &ffish);
+        babl_hash_table_find (id_htable, hashval, find_fish_path, (void *) &ffish);
 
         if (ffish.fish_path)
           {
@@ -279,57 +278,6 @@ babl_fish (const void *source,
   }
 }
 
-long
-babl_fish_process (Babl *babl,
-                   void *source,
-                   void *destination,
-                   long  n)
-{
-  long ret = 0;
-
-  switch (babl->class_type)
-    {
-      case BABL_FISH_REFERENCE:
-      case BABL_FISH_SIMPLE:
-      case BABL_FISH_PATH:
-        if (babl->class_type == BABL_FISH_REFERENCE)
-          {
-            if (babl->fish.source == babl->fish.destination)
-              { /* XXX: we're assuming linear buffers */
-                memcpy (destination, source, n * babl->fish.source->format.bytes_per_pixel);
-                ret = n;
-              }
-            else
-              {
-                ret = babl_fish_reference_process (babl, source, destination, n);
-              }
-          }
-        else if (babl->class_type == BABL_FISH_PATH)
-          {
-            ret = babl_fish_path_process (babl, source, destination, n);
-          }
-        else if (babl->class_type == BABL_FISH_SIMPLE)
-          {
-            if (BABL (babl->fish_simple.conversion)->class_type == BABL_CONVERSION_LINEAR)
-              {
-                ret = babl_conversion_process (BABL (babl->fish_simple.conversion),
-                                               source, destination, n);
-              }
-            else
-              {
-                babl_assert (0);
-              }
-          }
-        break;
-      default:
-        babl_log ("NYI");
-        ret = -1;
-        break;
-    }
-
-  return ret;
-}
-
 
 
 static int
